from enum import Enum
import numpy as np

class SearchSpaces(Enum):
    DECISION_TREE_CLASSIFIER = [{'estimator': ['decision_tree'],
                                'criterion': ['gini', 'entropy'],
                                'splitter': ['best', 'random'],
                                'max_depth': [3, 5, 8,10, 15],
                                'max_features': ['auto', 'sqrt', 'log2'],
                                'random_state': [42], }]

    XGB_CLASSIFIER =    [{'estimator': ['xgb'],
                        'n_estimators':[1500],
                        'use_label_encoder': [False],
                        'eta': list(np.linspace(0, 0.6, 3, dtype=float)),
                        'gamma':list(np.linspace(0, 0.3, 3,  dtype=float)),
                        'use_label_encoder': [False],
                        'n_estimators': [300, 500],
                        'eval_metric': ["logloss"],
                        'seed': [42],
                        'learning_rate': np.linspace(1e-5, 0.01, 3),
                        'gamma': np.linspace(0, 1, 2),
                        'max_depth':np.arange(3, 10, 2),
                        'min_child_weight': [1],
                        'subsample': np.linspace(0.5, 1, 3),
                        'colsample_bytree': np.linspace(0.5, 1, 3),
                        'alpha': np.linspace(1e-8, 10, 5),
                        'lambda': np.linspace(1e-8, 10, 5)}]

    
    CATBOOST_CLASSIFIER =  [{'estimator': ['catboost'],
                            'n_estimators': [500,1500],
                            'early_stopping_rounds': [50],
                            'eval_metric': ["AUC", "Logloss"],
                            'random_state': [42],
                            'learning_rate': np.linspace(1e-5, 0.01, 3),
                            'max_depth': [6, 8, 12],
                            'min_data_in_leaf': [20, 30, 50],
                            'colsample_bylevel': np.linspace(0.5, 1),
                            'boosting_type': ["Plain", "Ordered"],
                            'l2_leaf_reg': [1, 5, 7],
                            'loss_function': ['Logloss'],
                            'bootstrap_type':['Bayesian', 'Bernoulli', 'MVS'],
                            'multi_class':['ovr']}]


    ADABOOST_CLASSIFIER =   [{'estimator': ['adaboost'],
                                'n_estimators': [1000],
                                'algorithm': ["SAMME", "SAMME.R"],
                                'max_leaf_nodes': [16, 32]}]

    LGBM_CLASSIFIER = [{'estimator': ['lgbm'],
                        'objective': ['binary'],
                        'num_iterations': [1500],
                        'metric': ["auc", "logloss"],
                        'seed': [42],
                        'boosting': ['gbdt'],
                        'num_leaves':[16, 32, 64],
                        'learning_rate': [1e-5, 1e-3, 1e-2],
                        'max_depth': [2, 5, 7],
                        'min_data_in_leaf': [20, 50, 100],
                        'min_sum_hessian_in_leaf': [1e-5, 1e-3, 1e-2],
                        'lambda_l1': np.linspace(1e-3, 1, 3),
                        'lambda_l2': np.linspace(1e-3, 1, 3),
                        'bagging_freq': [5, 7],
                        'bagging_fraction': [0.2, 0.5, 1.0],
                        'feature_fraction': [0.6, 0.8, 1.0],
                        'feature_fraction_bynode': [0.6, 0.8, 1.0],
                        'min_data_per_group': [50, 100, 300],
                        'cat_l2': [20, 30],
                        'cat_smooth': [20, 30]}]

    LOGISTIC_REGRESSION = [{'estimator': ['logistic_regressor'],
                            # 'solver': ['lbfgs', 'liblinear', 'saga'],
                            'C': [1]
                            # 'C': [1, 0.8, 0.5, 0.3]}
                            }]


    DECISION_TREE_REGRESSOR = [{'estimator': ['decision_tree'],
                                'criterion': ['squared_error', 'friedman_mse', 'poisson'],
                                'splitter': ['best', 'random'],
                                'max_depth': [3, 5, 8, 10, 15],
                                'max_features': ['auto', 'sqrt', 'log2'],
                                'random_state': [42]}]

    XGB_REGRESSOR =    [{'estimator': ['xgb'],
                        'eta': list(np.linspace(0, 0.6, 3, dtype=float)),
                        'gamma':list(np.linspace(0, 0.3, 3,  dtype=float)),
                        'n_estimators': [300, 500],
                        'eval_metric': ["rmse"],
                        'seed': [42],
                        'learning_rate': np.linspace(1e-3, 0.01, 3),
                        'gamma': np.linspace(0, 0.5, 3),
                        'max_depth':np.arange(6, 15, 3),
                        'subsample': np.linspace(0.5, 1, 3),
                        'colsample_bytree': np.linspace(0, 1, 3),
                        'alpha': np.linspace(1e-8, 1, 3),
                        'lambda': np.linspace(1e-8, 1, 3)}]

   
    CATBOOST_REGRESSOR =  [{'estimator': ['catboost'],
                        'n_estimators': [1500],
                        'early_stopping_rounds': [50],
                        'eval_metric': ['RMSE'],
                        'random_state': [42],
                        'learning_rate': np.linspace(1e-5, 0.01, 3),
                        'max_depth': [7, 10, 12],
                        'min_data_in_leaf': [10, 30, 50],
                        'colsample_bylevel': [0.5, 1],
                        'boosting_type': ['Plain', 'Ordered'],
                        'l2_leaf_reg': [1, 5, 10],
                        'loss_function': ['RMSE'],
                        'bootstrap_type': ['Bayesian', 'Bernoulli', 'MVS']}]

    ADABOOST_REGRESSOR =   [{'estimator': ['adaboost'],
                            'n_estimators': [1000],
                            'loss': ["linear", "square", "exponenctial"],
                            'learning_rate': np.linspace(0.3, 1, 3), }]

    LGBM_REGRESSOR =    [{'estimator': ['lgbm'],
                        'objective': ['regression'],
                        'n_estimators': [1500],
                        'metric': ["rmse", "R2"],
                        'random_state': [42],
                        'boosting': ['gbdt'],
                        'num_leaves':[16, 32, 64],
                        'learning_rate': [1e-5, 1e-3, 1e-1],
                        'max_depth': [2, 5, 7],
                        'min_data_in_leaf': [10, 20, 50],
                        'min_sum_hessian_in_leaf': [1e-3, 1e-2, 1e-1],
                        'reg_alpha': np.linspace(1e-3, 1, 3),
                        'reg_lambda': np.linspace(1e-3, 1, 3),
                        'feature_fraction': [0.7, 0.8, 1.0],
                        'feature_fraction_bynode': [0.6, 0.8, 1.0],
                        'min_data_per_group': [100, 200],
                        'bagging_freq': [5, 7],
                        'bagging_fraction': [0.5, 0.7, 1.0],
                        'cat_l2': [10, 20, 50],
                        'categorical_feature': ['auto'],
                        'cat_smooth': [10, 20, 50]}]

    SVR  =           [{'estimator': ['svr'],
                    'kernel': ["linear", "poly", "sigmoid"],
                    'gamma':["scale", "auto"],
                    'C':[0.2, 0.5, 1],
                    'epsilon': [1e-2, 1-3, 1]}]

def get_enum(model, task):
    if task == 'classification':
        if model == 'decision_tree':
            return SearchSpaces.DECISION_TREE_CLASSIFIER.value
        elif model == 'xgb':
            return SearchSpaces.XGB_CLASSIFIER.value
        elif model == 'catboost':
            return SearchSpaces.CATBOOST_CLASSIFIER.value
        elif model == 'adaboost':
            return SearchSpaces.ADABOOST_CLASSIFIER.value
        elif model == 'lgbm':
            return SearchSpaces.LGBM_CLASSIFIER.value
        elif model == 'logistic_regression':
            return SearchSpaces.LOGISTIC_REGRESSION.value
        else:
            raise BaseException("Model not added to the tool yet")
    elif task == 'regression':
        if model == 'decision_tree':
            return SearchSpaces.DECISION_TREE_REGRESSOR.value
        elif model == 'xgb':
            return SearchSpaces.XGB_REGRESSOR.value
        elif model == 'catboost':
            return SearchSpaces.CATBOOST_REGRESSOR.value
        elif model == 'adaboost':
            return SearchSpaces.ADABOOST_REGRESSOR.value
        elif model == 'lgbm':
            return SearchSpaces.LGBM_REGRESSOR.value
        elif model == 'svr':
            return SearchSpaces.SVR.value
        else:
            raise BaseException("Model not added to the tool yet")
    else:
        raise BaseException("Task not implemented yet")
    





from .GridSearch import GridSearch
from .GridSearchCV import GridSearchCV
from  .OptunaSearch import OptunaSearch
#from .optuna_objectives import OptunaObjectives
from paml.searches.optuna_objectives import OptunaObjectives


import numpy as np
import pandas as pd
import logging


from sklearn.pipeline import Pipeline
from sklearn.model_selection import StratifiedKFold, KFold
from sklearn.model_selection import GridSearchCV as GridCV
from sklearn.ensemble import AdaBoostClassifier, AdaBoostRegressor
from sklearn.tree import DecisionTreeClassifier, DecisionTreeRegressor
from sklearn.linear_model import LogisticRegression
from sklearn.svm import SVR

from sklearn.metrics import roc_auc_score


from xgboost import XGBClassifier, XGBRegressor
from lightgbm import  LGBMClassifier, LGBMRegressor

from boruta import BorutaPy

from paml.searches import SearchSpaces
from paml.metrics import Metrics
from paml.searches.Base import CustomImputer, FeatureSelector, GetDummiesTransformer
from paml.searches.wrapper_catboost import CatBoostClassifierScikit, CatBoostRegressorScikit


class GridSearchCV:
    '''GridSearch implements fit, predict, get_best_model by KS (for classification only) '''
    def __init__(self, n_folds : int = 5,  verbose: int = 10, n_jobs: int = -1, 
                search_space: dict = None,  models:list=['all'], fs_params = {'max_depth': 5},
                task:str='classification', compute_ks: bool = False, acceptance_rate:float = 0.01,
                feature_selection:bool = False):

        self.search_space = search_space
        self.verbose = verbose
        self.n_jobs = n_jobs
        self.n_folds = n_folds
        self.models = models
        self.task = task
        self.compute_ks = compute_ks
        self.acceptance_rate = acceptance_rate
        self.feature_selection = feature_selection
        self.fs_params = fs_params     
       
        self.__set_model_params_grid__(search_space = self.search_space, models =  self.models, task = self.task)
                                                  
    def __retrieve_search_space_single_model__(self, models, search_space, task):    
            model_space = []
            for model in models:
                if model in ['xgb', 'catboost', 'adaboost', 'lgbm', 'logistic_regression', 'svr', 'decision_tree']:
                    for item in search_space:
                        if task == 'classification':
                            if item['clf'][0] == model:
                                item['clf'] = [self.get_estimator_instance_classifier(model)]
                                model_space.append(item)
                        elif task == 'regression':
                            if item['rgs'][0] == model:
                                item['rgs'] = [self.get_estimator_instance_regressor(model)]
                                model_space.append(item)
                        else: 
                            raise BaseException("This task is not implemented yet")
                else:
                    raise BaseException("This model is not in the AutoML yet")

                if not model_space:
                    for model in models:
                        search_item  = SearchSpaces.get_enum(model, task)
                        search_item = self.treat_search_space_keys(search_item)
                        search_item = self.set_estimator_instance(search_item)
                        if self.feature_selection:
                            search_item[0]['fs__params'] = [self.fs_params]
                        model_space.append(search_item[0])
                self.search_space = model_space

    def __set_model_params_grid__(self, search_space, models, task):
            if not search_space and models == ['all']:
                if task == 'classification':
                    models_enum = ['xgb', 'catboost', 'adaboost','lgbm', 'logistic_regression', 'decision_tree']
                    self.search_space = []
                    for model in models_enum:
                        search_item = SearchSpaces.get_enum(model, task)
                        search_item = self.treat_search_space_keys(search_item)
                        search_item = self.set_estimator_instance(search_item)
                        if self.feature_selection:
                            search_item[0]['fs__params'] = [self.fs_params]
                        self.search_space.append(search_item[0])
                elif task == 'regression':
                    models_enum = ['xgb', 'catboost', 'adaboost','lgbm', 'svr', 'decision_tree']
                    self.search_space = []
                    for model in models_enum:
                        search_item = SearchSpaces.get_enum(model, task)
                        search_item = self.treat_search_space_keys(search_item)
                        search_item = self.set_estimator_instance(search_item)
                        if self.feature_selection:
                            search_item[0]['fs__params'] = [self.fs_params]
                        self.search_space.append(search_item[0])                                      
                else:
                    raise BaseException("This task is not implemented yet")
            elif not search_space and models != ['all']:
                self.__set_search_space_single_models__(models, task)
            elif search_space and models != ['all']:
                self.search_space = self.treat_search_space_keys(search_space) 
                self.__retrieve_search_space_single_model__(models, self.search_space, self.task)
            else:
                pass

    def __set_search_space_single_models__(self, models, task):
        model_space = []
        for model in models:
            if model in ['xgb', 'catboost', 'adaboost', 'lgbm', 'logistic_regression', 'svr', 'decision_tree']:
                search_item = (SearchSpaces.get_enum(model, task))
                search_item = self.treat_search_space_keys(search_item)
                search_item = self.set_estimator_instance(search_item)
                if self.feature_selection:
                    search_item[0]['fs__params'] = [self.fs_params]
                model_space.append(search_item[0])         
            else:
                print("Model ", model, " is not in the AutoML yet")
        if not model_space:
            raise BaseException("None of the given models were added to this tool."+ 
            "The options are: 'xgb', 'catboost', 'adaboost', 'lgbm', 'logistic_regression', 'svr', 'decision_tree'")
        self.search_space = model_space

    def create_grid_classification(self):
        '''
            Creates a grid search according to constructor specs
        '''
        estimator_ = self.get_estimator_instance_classifier(self.models[0])
        

        if self.feature_selection:
            pipe = Pipeline(steps = [
                                    ('imputer', CustomImputer()),
                                    ('get_dummies', GetDummiesTransformer()),
                                    ('fs', FeatureSelector(self.fs_params)),
                                    ('clf', estimator_)])
        else:
            pipe = Pipeline(steps = [ ('imputer', CustomImputer()),
                                      ('get_dummies', GetDummiesTransformer()),
                                    ('clf', estimator_)])
                 
        cv = StratifiedKFold(n_splits = self.n_folds, shuffle=True, random_state=42)
        scoring = {"AUC": "roc_auc"}


        if not self.verbose:
             self.verbose = 10

        grid = GridCV(estimator= pipe,
                          param_grid = self.search_space,
                          cv = cv,
                          scoring = scoring,
                          return_train_score= True,
                          refit = "AUC",
                          verbose = self.verbose,
                          n_jobs =self.n_jobs)
        return grid

    def create_grid(self):
        if self.task == 'classification':
            return self.create_grid_classification()
        elif self.task == 'regression':
            return self.create_grid_regression()
        else:
            raise BaseException("Task not implemented yet")

    def set_estimator_instance(self, search_space):
        new_space = []
        for i, item in enumerate(search_space):
            new_space.append(item)
            if self.task == 'classification':
                new_space[i]['clf'] = [self.get_estimator_instance_classifier(new_space[i]['clf'][0])] 
            elif self.task == 'regression':
                new_space[i]['rgs'] = [self.get_estimator_instance_regressor(new_space[i]['rgs'][0])]
        return new_space  

    def get_estimator_instance_classifier(self, model):
        estimator = {}
        if  model == 'xgb':
            estimator = XGBClassifier(use_label_encoder=False)
        elif model == 'catboost' or model ==  'all':
            estimator = CatBoostClassifierScikit()
        elif model == 'adaboost':
            estimator = AdaBoostClassifier()
        elif model == 'lgbm':
            estimator = LGBMClassifier()
        elif model == 'decision_tree':
            estimator = DecisionTreeClassifier()
        elif model == 'logistic_regression':
            estimator = LogisticRegression()
        return estimator


    def get_estimator_instance_regressor(self, model):
        estimator = {}
        if model == 'xgb':
            estimator = XGBRegressor(label_encoder = False)
        elif model == 'catboost' or model == 'all':
            estimator = CatBoostRegressorScikit()
        elif model == 'adaboost':
            estimator = AdaBoostRegressor()
        elif model == 'lgbm':
            estimator = LGBMRegressor()
        elif model == 'decision_tree':
            estimator = DecisionTreeRegressor()
        elif model == 'svr':
            estimator = SVR()
        return estimator

    def create_grid_regression(self):
        '''
            Creates a grid search for regression according to the specs
        '''        
        self.imputer =  CustomImputer()               
        estimator = self.get_estimator_instance_regressor(self.models[0])

        if self.feature_selection:
            pipe = Pipeline(steps = [('imputer', CustomImputer()),
                            ('get_dummies', GetDummiesTransformer()),
                            ('fs', FeatureSelector(self.fs_params)),
                            ('rgs', estimator) ])
        else:
            pipe = Pipeline(steps = [('imputer',  CustomImputer()),
                                    ('get_dummies', GetDummiesTransformer()),
                                     ('rgs', estimator) ])

        cv = KFold(n_splits = self.n_folds, shuffle=True, random_state=42)
        theil_score = Metrics.theil_stat()
        scoring = {"theil": theil_score}

        # mse = make_scorer(score_func= mean_squared_error, needs_proba= False, greater_is_better=False)
        # scoring = {"neg_root_mean_squared_error": mse}


        if not self.verbose:
             self.verbose = 10

        grid = GridCV(estimator= pipe,
                          param_grid = self.search_space,
                          cv = cv,
                          scoring = scoring,
                          return_train_score= True,
                          refit = "theil",
                          verbose = self.verbose,
                          n_jobs =self.n_jobs)
        return grid

    def fit_model(self, X, y):
        '''
            Run a given grid search and returns the best model
        '''
        if type(X) == np.ndarray:
            X = pd.DataFrame(X)
        if type(y) == np.ndarray:
            y = pd.Series(y)
        


        best_model = []
        my_grid = self.create_grid()
        best_model = my_grid.fit(X,y)
        
        if self.task == 'classification':
            if self.compute_ks:
                best_model = self.get_best_model_by_ks(best_model, X, y)
                return best_model
            else:
                return best_model
        else:
             return best_model

    def get_best_model_by_ks(self, grid, X, y):

        grid_results = pd.DataFrame(grid.cv_results_)
        grid_results.sort_values(by='rank_test_AUC', inplace = True)
        num_models_accepted = int(np.ceil(self.acceptance_rate*len(grid_results)))
        best_models = ['']*(num_models_accepted)

        aucs = []
        cv = StratifiedKFold(n_splits = self.n_folds, shuffle=True, random_state=42)
        scoring = {"AUC": "roc_auc"}
        for params in grid_results.params:
            params = self.set_params_values_as_list(params)
            scoring = {"AUC": "roc_auc"}

            if self.feature_selection:
                pipe = Pipeline(steps = [('imputer', CustomImputer()),
                            ('get_dummies', GetDummiesTransformer()),
                            ('fs', BorutaPy(estimator = self.fs_estimator)),
                            ('clf', params['clf'][0]) ])
            else:
                pipe = Pipeline(steps = [('imputer',CustomImputer() ),
                        ('get_dummies', GetDummiesTransformer()),
                        ('clf', params['clf'][0]) ])

            model = GridCV(estimator= pipe,
                          param_grid = params,
                          cv = cv,
                          scoring = scoring,
                          return_train_score= True,
                          refit = "AUC",
                          verbose = self.verbose,
                          n_jobs =self.n_jobs)
            model.fit(X,y)
            y_pred = model.predict(X)
            auc = roc_auc_score(y, y_pred)
            aucs.append((auc, model))

        aucs.sort(key=lambda x: x[0], reverse=True)
        best_models = aucs[:num_models_accepted]

        model_ks = []
        for item in best_models:
            y_pred =  item[1].predict(X)
            ks = Metrics.__compute_ks_score__(y, y_pred)
            model_ks.append((ks, item[1]))

        model_ks.sort(key=lambda x: x[0], reverse=True)
        print('The model KS is: ', model_ks[0][0])
        return model_ks[0][1]

    def set_params_values_as_list(self, params):
        new_params = {}
        for key,value in params.items():
            new_params[key] =[value]

        return new_params

    def parametrize_model(self, params, model):
        if isinstance(model, CatBoostClassifierScikit):
            return CatBoostClassifierScikit(params)
        elif isinstance(model,CatBoostRegressorScikit):
            return CatBoostRegressorScikit(params)
        elif isinstance(model, XGBClassifier):
            return XGBClassifier(**params)
        elif isinstance(model, XGBRegressor):
            return XGBRegressor(**params)
        elif isinstance(model, AdaBoostClassifier):
            return AdaBoostClassifier(**params)
        elif isinstance(model, AdaBoostRegressor):
            return AdaBoostRegressor(**params)
        elif isinstance(model, LGBMClassifier):
            return LGBMClassifier(**params)
        elif isinstance(model, LGBMRegressor):
            return LGBMRegressor(**params)
        elif isinstance(model, DecisionTreeClassifier):
            return DecisionTreeClassifier(**params)
        elif isinstance(model, DecisionTreeRegressor):
            return DecisionTreeRegressor(**params)
        elif isinstance(model, LogisticRegression):
            return LogisticRegression(**params)
        elif isinstance(model, SVR):
            return SVR(**params)
        else:
            raise BaseException ("Model not implemented yet")

    def treat_search_space_keys(self, search_space):
        new_space = []
        for i, item in enumerate(search_space):
            new_space.append({})
            for key, value in item.items():
                if key != 'estimator' and key != 'fs__params':
                    if self.task == 'classification':
                        new_space[i]['clf__'+key] = value
                    elif self.task == 'regression':
                        new_space[i]['rgs__'+key] = value
                else:
                    if key != 'fs__params':
                        if self.task == 'classification':
                            new_space[i]['clf'] = value
                        elif self.task == 'regression':
                            new_space[i]['rgs'] = value
        return new_space   



import numpy as np
from matplotlib import pyplot as plt

import pandas as pd
from sklearn.metrics import roc_auc_score
from sklearn.model_selection import StratifiedKFold, KFold, ParameterGrid

from sklearn.pipeline import Pipeline
from sklearn.model_selection import cross_validate
from sklearn.ensemble import RandomForestRegressor, AdaBoostClassifier, AdaBoostRegressor
from sklearn.tree import DecisionTreeClassifier, DecisionTreeRegressor
from sklearn.linear_model import LogisticRegression
from sklearn.svm import SVR

from xgboost import XGBClassifier, XGBRegressor
from lightgbm import LGBMClassifier, LGBMRegressor

from sklearn.datasets import make_blobs, make_classification, make_regression

import optuna

from paml.metrics import Metrics
from paml.searches import Base
from paml.searches.optuna_objectives import OptunaObjectives
from paml.searches.Base import CustomImputer
from paml.searches.wrapper_catboost import CatBoostClassifierScikit, CatBoostRegressorScikit

class OptunaSearch:
    '''OptunaSearch implements fit, predict (from the estimator) get_best_model, get_best_model_by_ks (for classification only)'''
    def __init__(self, task:str, models:str = ['all'], n_trials:int = 10, n_folds:int = 5,
    compute_ks:bool = False,
    feature_selection:bool = False, multi_class = False):
        self.models = models
        self.n_trials = n_trials
        self.task = task
        self.compute_ks = compute_ks
        self.n_folds = n_folds
        self.feature_selection = feature_selection
        self.multi_class = multi_class
        

        self.feature_selector_search_space = [{'max_depth': [5]}]
        
        if self.models == ['all']:
            if self.task == 'classification':
                self.objectives = {'xgb' : OptunaObjectives.__objective_xgb_clf__,
                                    'catboost' : OptunaObjectives.__objective_catboost_classifier__,
                                    'lgbm': OptunaObjectives.__objective_lgbm_classifier__,
                                    'decision_tree': OptunaObjectives.__objective_decision_tree_classifier__,
                                    'adaboost': OptunaObjectives.__objective_adaboost_classifier__,
                                    'logistic_regression': OptunaObjectives.__objective__logistic_regression__}
            elif self.task == 'regression':
                self.objectives = {'xgb':OptunaObjectives.__objective_xgb_regressor__, 
                                   'catboost':OptunaObjectives.__objective_catboost_regressor__,
                                   'lgbm': OptunaObjectives.__objective_lgbm_regressor__,
                                   'decision_tree': OptunaObjectives.__objective_decision_tree_regressor__,
                                   'adaboost': OptunaObjectives.__objective_adaboost_regressor__,
                                   'svr': OptunaObjectives.__objective_svr_regressor__}
            else:
                raise BaseException('Task type not implemented yet')
        else:
            self.__set_model_objective_function__()
            
        if self.compute_ks and self.task == 'regression':
            raise BaseException("You can't compute KS for regression tasks, please set compute_ks to False")
    
    def __set_model_objective_function__(self):
        self.objectives = {}
        for model in self.models:
            if model == 'xgb':
                if self.task == 'classification':
                    self.objectives['xgb'] = None
                    self.objectives['xgb'] = OptunaObjectives.__objective_xgb_clf__
                elif self.task == 'regression':
                    self.objectives['xgb'] =  OptunaObjectives.__objective_xgb_regressor__
            elif model == 'catboost':
                if self.task == 'classification':
                    self.objectives['catboost'] = None
                    if self.multi_class:
                        self.objectives['catboost'] = OptunaObjectives.__objective_catboost_classifier_multi__
                    else:
                        self.objectives['catboost'] = OptunaObjectives.__objective_catboost_classifier__

                elif self.task == 'regression':
                    self.objectives['catboost'] = OptunaObjectives.__objective_catboost_regressor__        
            elif model == 'lgbm':
                if self.task == 'classification':
                    self.objectives['lgbm'] = None
                    if self.multi_class:
                        self.objectives['lgbm'] = OptunaObjectives.__objective_lgbm_classifier_multi__
                    else:
                        self.objectives['lgbm'] = OptunaObjectives.__objective_lgbm_classifier__

                elif self.task == 'regression':
                    self.objectives['lgbm'] =  OptunaObjectives.__objective_lgbm_regressor__
            elif model == 'decision_tree':
                if self.task == 'classification':
                    self.objectives['decision_tree'] = OptunaObjectives.__objective_decision_tree_classifier__
                elif self.task == 'regression':
                    self.objectives['decision_tree'] = OptunaObjectives.__objective_decision_tree_regressor__
            elif model == 'adaboost':
                if self.task == 'classification':
                    self.objectives['adaboost'] = OptunaObjectives.__objective_adaboost_classifier__
                elif self.task == 'regression':
                    self.objectives['adaboost'] = OptunaObjectives.__objective_adaboost_regressor__
            elif model == 'svr':
                self.objectives['svr'] = OptunaObjectives.__objective_svr_regressor__
            elif model == 'logistic_regression':
                self.objectives ['logistic_regression'] = OptunaObjectives.__objective__logistic_regression__
            else:
                raise BaseException('This model was not added to the tool yet')
                
    def __tune_params__(self, X, y, objective, direction, n_trials):
        objective_lambda = lambda trial: objective(trial, X, y)
        study = optuna.create_study(study_name = 'OptunaStudy', direction = direction)
        study.optimize(objective_lambda, n_trials = n_trials, n_jobs=-1)
        return study.best_params
    

    def __get_internal_best_model__(self, models_per_objective):
        best_metric = models_per_objective[0][0]
        best_model = models_per_objective[0][1]
        if self.task == 'classification':
            for item in models_per_objective:
                if item[0] > best_metric:
                    best_metric = item[0]
                    best_model = item[1]
        elif self.task == 'regression':
            for item in models_per_objective:
                if item[0] < best_metric:
                    best_metric = item[0]
                    best_model = item[1]
        return best_model
    
    def __get_best_model_by_simple_metric__(self, results):
        best_metric = results[0][0]
        best_model = results[0][1]
        if self.task == 'classification':
            for item in results:
                if item[0] > best_metric:
                    best_metric = item[0]
                    best_model = item[1]
        elif self.task == 'regression':
            for item in results:
                if item[0] < best_metric:
                    best_metric = item[0]
                    best_model = item[1]
        
        print("Best score: ", best_metric)
        return best_model
            
        
    def __get_best_model_by_ks__(self, ks_results):
        best_model = ks_results[0][1]
        best_ks = ks_results[0][0]
        for item in ks_results:
            if item[0] > best_ks:
                best_ks = item[0]
                best_model = item[1]
        print("KS: ", best_ks)
        return best_model
        
    def __fit_models__(self, objectives, X, y, task):
        models = []
        if type(X) == np.ndarray:
            X = pd.DataFrame(X)
        if type(y) == np.ndarray:
            y = pd.Series(y)
        best_models = []
        
       
        imputer = CustomImputer()
        X = pd.DataFrame(imputer.fit_transform(X), columns=X.columns)

        self.feature_selector_settings = list(ParameterGrid(self.feature_selector_search_space))
        
        if task == 'classification':
            kfold =  StratifiedKFold(n_splits = self.n_folds)  
            for model_name, obj_function in objectives.items():
                pred = []
                pipe = None
                models_per_objective = []
                for train_idx, test_idx in kfold.split(X, y):
                    params = self.__tune_params__(X.iloc[train_idx,:], y.iloc[train_idx], obj_function, "maximize", self.n_trials)
                    if self.feature_selection:
                        pipe = self.get_pipe_with_feature_selection(model_name, self.task, params)
                    else:
                        pipe = self.get_pipe(model_name, self.task, params)

                    pipe.fit(X.iloc[train_idx,:], y.iloc[train_idx])
                    pred = None

                    metric = None
                    if len(np.unique(y)) > 2:
                        pred = pipe.predict_proba(X.iloc[test_idx,:])          
                        metric = roc_auc_score(y.iloc[test_idx], pred,multi_class= 'ovo',   average='weighted')
                    else:
                        pred = pipe.predict(X.iloc[test_idx,:])          
                        metric = roc_auc_score(y.iloc[test_idx], pred)
          
                    models_per_objective.append((metric, pipe))     
                best_models.append(self.__get_internal_best_model__(models_per_objective))
                
        elif task == 'regression':
            kfold =  KFold(n_splits = self.n_folds)
            for model_name, obj_function in objectives.items():
                print(model_name)
                pred = []
                pipe = None
                models_per_objective = []
                for train_idx, test_idx in kfold.split(X, y):
                    params = self.__tune_params__(X.iloc[train_idx,:], y.iloc[train_idx], obj_function, "minimize", self.n_trials)
                    if self.feature_selection:
                        pipe = self.get_pipe_with_feature_selection(model_name, self.task, params)
                    else:
                        pipe = self.get_pipe(model_name, task, params)
                    pipe.fit(X.iloc[train_idx,:], y.iloc[train_idx])
                    pred = pipe.predict(X.iloc[test_idx,:])
                    
                    metric = Metrics.__theil_stat__(y.iloc[test_idx], pred)
                    models_per_objective.append((metric, pipe))
                best_models.append(self.__get_internal_best_model__(models_per_objective))
        return best_models
    
    def get_pipe_with_feature_selection(self, model_name, task, params):
        for fs_settings in self.feature_selector_settings:
            if task == 'classification':
                if model_name == 'xgb':
                        pipe = Pipeline(steps = [('imputer', CustomImputer()),
                                                ("dummies", Base.GetDummiesTransformer()),
                                                ('fs', Base.FeatureSelector(fs_settings)),
                                                ('estimator', Base.Estimator(XGBClassifier(**params)))])
                elif model_name == 'catboost':
                    pipe = Pipeline(steps = [('imputer', CustomImputer()),
                                            ('fs', Base.FeatureSelector(fs_settings)),
                                            ('estimator', Base.Estimator(CatBoostClassifierScikit(params)))])                   
                elif model_name == 'lgbm':
                    pipe = Pipeline(steps = [('imputer', CustomImputer()),
                                            ("dummies", Base.GetDummiesTransformer()),
                                            ('fs', Base.FeatureSelector(fs_settings)),
                                            ('estimator', Base.Estimator(LGBMClassifier(**params)))]) 
                elif model_name == 'decision_tree':
                    pipe = Pipeline(steps = [('imputer',CustomImputer()),
                                            ("dummies", Base.GetDummiesTransformer()),
                                            ('fs', Base.FeatureSelector(fs_settings)),
                                            ('estimator', Base.Estimator(DecisionTreeClassifier(**params)))]) 
                elif model_name == 'adaboost':
                    pipe = Pipeline(steps = [('imputer', CustomImputer()),
                                            ("dummies", Base.GetDummiesTransformer()),
                                            ('fs', Base.FeatureSelector(fs_settings)),
                                            ('estimator', Base.Estimator(AdaBoostClassifier(**params)))])
                elif model_name == 'logistic_regression':       
                    pipe = Pipeline(steps = [('imputer',CustomImputer()),
                            ('fs', Base.FeatureSelector(fs_settings)),
                            ('estimator', Base.Estimator(LogisticRegression(**params)))])    
                else:
                    raise BaseException("Model not added to the tool yet")

            elif task == 'regression':
                if model_name == 'xgb':
                        pipe = Pipeline(steps = [('imputer', CustomImputer()),
                                                ("dummies", Base.GetDummiesTransformer()),
                                                ('fs', Base.FeatureSelector(fs_settings)),
                                                ('estimator', Base.Estimator(XGBRegressor(**params)))])
                elif model_name == 'catboost':
                    pipe = Pipeline(steps = [('imputer', CustomImputer()),
                                            ('fs', Base.FeatureSelector(fs_settings)),
                                            ('estimator', Base.Estimator(CatBoostRegressorScikit(params)))])                   
                elif model_name == 'lgbm':
                    pipe = Pipeline(steps = [('imputer', CustomImputer()),
                                            ("dummies", Base.GetDummiesTransformer()),
                                            ('fs', Base.FeatureSelector(fs_settings)),
                                            ('estimator', Base.Estimator(LGBMRegressor(**params)))]) 
                elif model_name == 'decision_tree':
                    pipe = Pipeline(steps = [('imputer', CustomImputer()),
                                            ("dummies", Base.GetDummiesTransformer()),
                                            ('fs', Base.FeatureSelector(fs_settings)),
                                            ('estimator', Base.Estimator(DecisionTreeRegressor(**params)))]) 
                elif model_name == 'adaboost':
                    pipe = Pipeline(steps = [('imputer', CustomImputer()),
                                            ("dummies", Base.GetDummiesTransformer()),
                                            ('fs', Base.FeatureSelector(fs_settings)),
                                            ('estimator', Base.Estimator(AdaBoostRegressor(**params)))])
                elif model_name == 'svr':       
                    pipe = Pipeline(steps = [('imputer', CustomImputer()),
                                            ("dummies", Base.GetDummiesTransformer()),
                                            ('fs', Base.FeatureSelector(fs_settings)),
                                            ('estimator', Base.Estimator(SVR(**params)))])    
                else:
                    raise BaseException("Model not added to the tool yet")
        return pipe

    def get_pipe(self, model_name, task, params):
        if task == 'classification':
            if model_name == 'xgb':
                    pipe = Pipeline(steps = [('imputer', CustomImputer()),
                                            ("dummies", Base.GetDummiesTransformer()),
                                            ('estimator', Base.Estimator(XGBClassifier(**params)))])
            elif model_name == 'catboost':
                pipe = Pipeline(steps = [('imputer', CustomImputer()),
                                        ('estimator', Base.Estimator(CatBoostClassifierScikit(params)))])                   
            elif model_name == 'lgbm':
                pipe = Pipeline(steps = [('imputer',CustomImputer()),
                                        ("dummies", Base.GetDummiesTransformer()),
                                        ('estimator', Base.Estimator(LGBMClassifier(**params)))]) 
            elif model_name == 'decision_tree':
                pipe = Pipeline(steps = [('imputer', CustomImputer()),
                                        ("dummies", Base.GetDummiesTransformer()),
                                        ('estimator', Base.Estimator(DecisionTreeClassifier(**params)))]) 
            elif model_name == 'adaboost':
                pipe = Pipeline(steps = [('imputer', CustomImputer()),
                                        ("dummies", Base.GetDummiesTransformer()),
                                        ('estimator', Base.Estimator(AdaBoostClassifier(**params)))])
            elif model_name == 'logistic_regression':       
                pipe = Pipeline(steps = [('imputer', CustomImputer()),
                                        ("dummies", Base.GetDummiesTransformer()),
                                        ('estimator', Base.Estimator(LogisticRegression(**params)))])    
            else:
                raise BaseException("Model not added to the tool yet")
        elif task == 'regression':
            if model_name == 'xgb':
                    pipe = Pipeline(steps = [('imputer', CustomImputer()),
                                            ("dummies", Base.GetDummiesTransformer()),
                                            ('estimator', Base.Estimator(XGBRegressor(**params)))])
            elif model_name == 'catboost':
                pipe = Pipeline(steps = [('imputer', CustomImputer()),
                                        ('estimator', Base.Estimator(CatBoostRegressorScikit(params)))])                   
            elif model_name == 'lgbm':
                pipe = Pipeline(steps = [('imputer', CustomImputer()),
                                        ("dummies", Base.GetDummiesTransformer()),
                                        ('estimator', Base.Estimator(LGBMRegressor(**params)))]) 
            elif model_name == 'decision_tree':
                pipe = Pipeline(steps = [('imputer', CustomImputer()),
                                        ("dummies", Base.GetDummiesTransformer()),
                                        ('estimator', Base.Estimator(DecisionTreeRegressor(**params)))]) 
            elif model_name == 'adaboost':
                pipe = Pipeline(steps = [('imputer', CustomImputer()),
                                        ("dummies", Base.GetDummiesTransformer()),
                                        ('estimator', Base.Estimator(AdaBoostRegressor(**params)))])
            elif model_name == 'svr':       
                pipe = Pipeline(steps = [('imputer', CustomImputer()),
                                        ("dummies", Base.GetDummiesTransformer()),
                                        ('estimator', Base.Estimator(SVR(**params)))])    
            else:
                raise BaseException("Model not added to the tool yet")
        return pipe
      
    def __get_best_model__(self, X, y, task, ks_flag, models):
        if ks_flag and task == 'classification':
            ks_results = []
            for model in models:
                model.fit(X, y)
                y_pred = model.predict(X)
                ks = Metrics.__compute_ks_score__(y, y_pred)
                ks_results.append((ks, model))  
            return self.__get_best_model_by_ks__(ks_results)

        if not ks_flag and task == 'classification':
            roc_results = []
            for model in models:
                model.fit(X, y)
                y_pred = None
                roc = None

                if len(np.unique(y)) > 2:
                    y_pred = model.predict_proba(X)
                    roc = roc_auc_score(y, y_pred, multi_class='ovr', average='weighted')
                else:
                    y_pred = model.predict(X)
                    roc = roc_auc_score(y, y_pred)


                roc_results.append((roc, model))
            return self.__get_best_model_by_simple_metric__(roc_results)     
        elif task == 'regression':
            results = []
            for model in models:
                model.fit(X, y)
                y_pred = model.predict(X)
                theil_stat = Metrics.__theil_stat__(y, y_pred)
                results.append((theil_stat, model))
            return self.__get_best_model_by_simple_metric__(results)
        else:
            raise BaseException("Task not implemented yet")
   
    def fit(self, X, y):           
        models = self.__fit_models__(self.objectives, X, y, self.task)
        return self.__get_best_model__(X, y, self.task, self.compute_ks, models)


import numpy as np
import pandas as pd
from pandas.api.types import is_numeric_dtype

from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.ensemble import RandomForestRegressor

from boruta import BorutaPy
from paml.searches.wrapper_catboost import CatBoostClassifierScikit, CatBoostRegressorScikit
import warnings


class FeatureSelector(TransformerMixin):
    '''Custom feature selection class using BorutaPy'''
    def __init__(self, params):
        self.params = params

    def fit(self, X, y):
        self.accept = []
        self.selector = BorutaPy(estimator=RandomForestRegressor(max_depth=self.params['max_depth'], n_jobs=-1),
                                 max_iter=200, n_estimators='auto')

        numeric_features = [col for col in X.columns if is_numeric_dtype(X[col]) or np.all(X[col].str.isnumeric())]
        if numeric_features:
            X_boruta = X[numeric_features]

            self.selector.fit(X_boruta.values, y)
            self.accept = X_boruta.columns[self.selector.support_].to_list()
        return self

    def transform(self, X, y=None):
        cat_features = [col for col in X.columns if X[col].dtypes  == 'object']      
        final_df = None
        # cat_features_index = [X.columns.get_loc(x) for x in X.columns if X[x].dtypes == 'object']
        if self.accept:
            if cat_features:
                final_df = X.loc[:, self.accept]
                final_df[cat_features] = X[cat_features].copy()
            else:
                final_df = X.loc[:, self.accept] 
            return final_df
        else:
            return X


class Estimator(BaseEstimator):
    warnings.filterwarnings("ignore")

    '''Custom Estimator Class for fitting and predicting models on Pipelines'''
    def __init__(self, model):
        self.model = model

    def fit(self, X, y=None):
        if type(X) == np.ndarray:
            X = pd.DataFrame(X)
        # if not (isinstance(self.model, CatBoostClassifierScikit) or isinstance(self.model, CatBoostRegressorScikit)):
        #     X = pd.get_dummies(data = X, drop_first = True)
        return self.model.fit(X, y)

    def predict(self, X):
    
        if type(X) == np.ndarray:
                X = pd.DataFrame(X)
                
        numeric_features = [col for col in X.columns if is_numeric_dtype(X[col]) or np.all(X[col].str.isnumeric())]

        if numeric_features:
            for col in X[numeric_features].columns.copy():
            #    X.loc[:][col] = X.loc[:,col].astype('float')
                X.loc[:, col] = X[col].copy().astype('float64')


        if not (isinstance(self.model, CatBoostClassifierScikit) or isinstance(self.model, CatBoostRegressorScikit)):
            X = pd.get_dummies(X, drop_first=True)
        return self.model.predict(X)

    def predict_proba(self, X):
        return self.model.predict_proba(X)

    
    def get_estimator(self):
        return self.model


class CustomImputer:

    '''Custom Imputation class for imputing missing data, imputes mean for real data and most frequent 
    categorical data'''
    def __init__(self):
        pass
    def fit( self, X, y = None):
        if type(X) == np.ndarray:
            X = pd.DataFrame(X)
            X = X.convert_dtypes(convert_integer=True, convert_string = False)
            numeric_features_index = [X.columns.get_loc(col) for col in X.columns if is_numeric_dtype(X[col]) or np.all(X[col].str.isnumeric())]
            if numeric_features_index:
                for feat in X.columns:
                    if feat in X.columns[numeric_features_index]:
                        # X.loc[:][feat] = X[feat].astype('float64')
                        X.loc[:, feat] = X[feat].copy().astype('float64')





        self.fill_values = pd.Series([X[column].value_counts().index[0] 
                                if X[column].dtype == np.dtype('O') and  X[column].isnull().sum() == len(X[column])
                                else  0 if X[column].isnull().sum() == len(X[column]) 
                                        else X[column].mean() if X[column].dtype != np.dtype('O') else X[column].value_counts().index[0] 
                                            for column in X.columns], index = X.columns)

        return self

    def transform(self, X, y=None):
        if type(X) == np.ndarray:
            X = pd.DataFrame(X)
            X = X.convert_dtypes(convert_integer=True, convert_string = False)
        numeric_features_index = [X.columns.get_loc(x) for x in X.columns if is_numeric_dtype(X[x]) or np.all(X[x].str.isnumeric())]
        if numeric_features_index:
            for feat in X.columns.copy().values:
                if feat in X.columns[numeric_features_index].copy():
                    X.loc[:, feat] = X[feat].copy().astype('float64')
                    # X.loc[:][feat] = X[feat].astype('float64')

        return X.fillna(self.fill_values)

    def fit_transform(self, X, y = None):
        self.fit(X)
        return self.transform(X)


class GetDummiesTransformer(TransformerMixin):  
    def __init__(self) -> None:
        self.training_columns = None
        pass
    def fit(self, X, y=None):
        return self
    def transform(self, X, y = None):
       
        cats = [col for col in X.columns if X[col].dtypes == 'object']
        X = pd.get_dummies(X, drop_first=True, columns = cats)
        if not np.any(self.training_columns):
            self.training_columns = X.columns
        else:
            missing_cols = set(self.training_columns) - set(X.columns )

            for col in missing_cols:
                X[col] = 0
            X = X[self.training_columns]
        
        return X


from catboost import CatBoostClassifier, CatBoostRegressor, Pool
from pandas.api.types import is_numeric_dtype
import pandas as pd
import numpy as np

class CatBoostClassifierScikit(CatBoostClassifier):
    def __init__(self, params={}) -> None:
        self.classifier = CatBoostClassifier(**params)

    def fit(self, X, y):
        if type(X) == np.ndarray:
            X = pd.DataFrame(X)

        cat_index = [X.columns.get_loc(
            x) for x in X.columns if X[x].dtypes == 'object']
        numeric_features_index = [X.columns.get_loc(x) for x in X.columns if is_numeric_dtype(X[x])]
        for feat in X.columns:
            if feat in X.columns[numeric_features_index]:
                X.loc[:, feat] = X[feat].copy().astype('float64')

        pool  = Pool(data=X[X.columns], label=y, cat_features=cat_index)
        self.classifier.fit(pool)

        return self

    def predict(self, X, y=None):
        if type(X) == np.ndarray:
            X = pd.DataFrame(X)

        cat_index = [X.columns.get_loc(
            x) for x in X.columns if X[x].dtypes == 'object']
        numeric_features_index = [X.columns.get_loc(x) for x in X.columns if is_numeric_dtype(X[x])]
        for feat in X.columns:
            if feat in X.columns[numeric_features_index]:
                X[feat] = X[feat].astype('float')
        pool  = Pool(data=X[X.columns], label=y, cat_features=cat_index)
        return self.classifier.predict(pool)


    def predict_proba(self, X, y=None):
        if type(X) == np.ndarray:
            X = pd.DataFrame(X)

        cat_index = [X.columns.get_loc(
            x) for x in X.columns if X[x].dtypes == 'object']
        numeric_features_index = [X.columns.get_loc(x) for x in X.columns if is_numeric_dtype(X[x])]
        for feat in X.columns:
            if feat in X.columns[numeric_features_index]:
                X[feat] = X[feat].astype('float')
        pool  = Pool(data=X[X.columns], label=y, cat_features=cat_index)
        return self.classifier.predict_proba(pool)

    def set_params(self, **params):
        self.classifier.set_params(**params)

    
       
class CatBoostRegressorScikit:
    def __init__(self, params={}):
        self.regressor = CatBoostRegressor(**params)
        # super().__init__(loss_function=None)

    def fit(self, X, y):
        if type(X) == np.ndarray:
            X = pd.DataFrame(X)

        cat_index = [X.columns.get_loc(
            x) for x in X.columns if X[x].dtypes == 'object']
        numeric_features_index = [X.columns.get_loc(x) for x in X.columns if is_numeric_dtype(X[x])]
        for feat in X.columns:
            if feat in X.columns[numeric_features_index]:
                X[feat] = X[feat].astype('float')
        pool  = Pool(data=X[X.columns], label=y, cat_features=cat_index)
        self.regressor.fit(pool)

        return self

    def predict(self, X, y=None):
        if type(X) == np.ndarray:
            X = pd.DataFrame(X)

        cat_index = [X.columns.get_loc(
            x) for x in X.columns if X[x].dtypes == 'object']
        numeric_features_index = [X.columns.get_loc(x) for x in X.columns if is_numeric_dtype(X[x])]
        for feat in X.columns:
            if feat in X.columns[numeric_features_index]:
                X[feat] = X[feat].astype('float')
        pool  = Pool(data=X[X.columns], label=y, cat_features=cat_index)
        return self.regressor.predict(pool)


    def set_params(self, **params):
        self.regressor.set_params(**params)

    @property
    def boosting_type(self, boosting_type):
        self.regressor.boosting_type = boosting_type


import copy
from re import M
import numpy as np
import pandas as pd
from sklearn.pipeline import Pipeline
from sklearn.metrics import roc_auc_score
from sklearn.model_selection import train_test_split
from sklearn.model_selection import ParameterGrid
from sklearn.ensemble import RandomForestRegressor, AdaBoostClassifier, AdaBoostRegressor
from sklearn.tree import DecisionTreeClassifier, DecisionTreeRegressor
from sklearn.linear_model import LogisticRegression
from sklearn.svm import SVR

from xgboost import XGBClassifier, XGBRegressor
from lightgbm import LGBMClassifier, LGBMRegressor

from boruta import BorutaPy
from multiprocessing.dummy import Pool as ThreadPool

from paml.metrics import Metrics
from paml.searches import Base
from paml.searches import SearchSpaces
from paml.searches.wrapper_catboost import CatBoostClassifierScikit, CatBoostRegressorScikit

class GridSearch:
    '''GridSearch implements fit, get_best_model, get_best_model by KS (for classification only), predict (from the estimator) '''
    def __init__(self, task: str,
                 search_space: dict = None,
                 feature_selector_search_space=None,
                 models=['all'],
                 compute_ks: bool = False,
                 acceptance_rate: float = 0.01,
                 feature_selection: bool = False):

        self.search_space = search_space
        self.feature_selector_search_space = feature_selector_search_space
        self.models = models
        self.task = task
        self.compute_ks = compute_ks
        self.acceptance_rate = acceptance_rate
        self.feature_selection = feature_selection

        if self.task != 'classification' and self.compute_ks == True:
            raise BaseException(
                "You can't use compute_ks as True in this taks. Please set it to False")

        self.__set_model_params_grid__(
            self.search_space, self.models, self.task)

        if not self.feature_selector_search_space:
            self.feature_selector_search_space = [{'selector': [BorutaPy(estimator=RandomForestRegressor(n_jobs=-1))],
                                                   'max_depth': [5]}]

    def __set_model_params_grid__(self, search_space, models, task):
        if not search_space and models == ['all']:
            if task == 'classification':
                models_enum = ['xgb', 'catboost', 'adaboost','lgbm', 'logistic_regression', 'decision_tree']
                self.search_space = []
                for model in models_enum:
                    self.search_space.append(SearchSpaces.get_enum(model, task)[0])
            elif task == 'regression':
                models_enum = ['xgb', 'catboost', 'adaboost','lgbm', 'svr', 'decision_tree']
                for model in models_enum:
                    self.search_space.append(SearchSpaces.get_enum(model, task)[0])
            else:
                raise BaseException("This task in not implemented yet")
        elif not search_space and models != ['all']:
            self.__set_search_space_single_models__(models, task)
        elif search_space and models != ['all']:
            self.__retrieve_search_space_single_model__(
                models, self.search_space, self.task)
        elif search_space and models == ['all']:
            pass

    def __retrieve_search_space_single_model__(self, models, search_space, task):
        model_space = []
        for model in models:
            if model in ['xgb', 'catboost', 'adaboost', 'lgbm', 'logistic_regression', 'svr', 'decision_tree']:
                for item in search_space:
                    if task == 'classification':
                        if item['estimator'][0] == model:
                            model_space.append(item)
                    elif task == 'regression':
                        if item['estimator'][0] == model:
                            model_space.append(item)
            else:
                raise BaseException("This model is not in the AutoML yet")
        if not model_space:
            for model in models:
                item = SearchSpaces.get_enum(model, task)[0]
                model_space.append(item)
        self.search_space = model_space

    def __set_search_space_single_models__(self, models, task):
        model_space = []
        for model in models:
            if model in ['xgb', 'catboost', 'adaboost', 'lgbm', 'logistic_regression', 'svr', 'decision_tree']:
                model_space.append(SearchSpaces.get_enum(model, task)[0])
            else:
                print("Model ", model, " is not in the AutoML yet")
        if not model_space:
            raise BaseException("None of the given models were added to this tool."+ 
            "The options are: 'xgb', 'catboost', 'adaboost', 'lgbm', 'logistic_regression', 'svr', 'decision_tree'")
        self.search_space = model_space



    def fit(self, X, y):
        if type(X) == np.ndarray:
            X = pd.DataFrame(X)
        if type(y) == np.ndarray:
            y = pd.Series(y)
            
        if self.task == 'classification':
            return self.run_grid_search_classification(X, y)
        elif self.task == 'regression':
            return self.run_grid_search_regression(X, y)
        else:
            raise BaseException("This task was not implemented yet")


    
    def run_grid_search_classification(self, X, y):
        ''' Runs a pipe with all the giveng settings'''

        X_train, X_test, y_train, y_test = train_test_split(
            X, y, test_size=0.2, stratify=y)

        param_grid = list(ParameterGrid(self.search_space))
        param_grid_selector = list(ParameterGrid(
        self.feature_selector_search_space))
        param_grid_selector_cp = copy.deepcopy(param_grid_selector)
        parameter_grid_cp = copy.deepcopy(param_grid)


        models = self.__run_pipe__(
            X_train, y_train, parameter_grid_cp, param_grid_selector_cp)
        best_model = self.__get_best_model__(self.task, models, X_test, y_test)

        best_model = best_model.fit(X, y)
        return best_model

    def run_grid_search_regression(self, X, y):
        X_train, X_test, y_train, y_test = train_test_split(
            X, y, test_size=0.2)

        param_grid = list(ParameterGrid(self.search_space))

        param_grid_selector = list(ParameterGrid(
            self.feature_selector_search_space))

        param_grid_selector_cp = copy.deepcopy(param_grid_selector)

        models = self.__run_pipe__(
            X_train, y_train, param_grid, param_grid_selector_cp)
        best_model = self.__get_best_model__(
            self.task, models, X_test, y_test)

        best_model = best_model.fit(X, y)
        return best_model

    def __run_pipe__(self, X_train, y_train, model_space_params, feature_selector_space_params):
        '''
            Runs the complete pipe for every setting informed and returns the list of fitted models
        '''
        models = []
        if self.feature_selection:
            models = self.run_pipe_with_fs(model_space_params,feature_selector_space_params, X_train, y_train )  
        else:    
            models = self.run_pipe_with_model_params(model_space_params, X_train, y_train)
        return models


    def run_pipe_with_model_params(self, model_params, X_train, y_train):
        '''Creates, Run, and stores a full pipe for each possible parameter setting'''
        models = []
        for i, item in enumerate(model_params):  
            print("Fit ", i+1, "/ " + str(len(model_params)))
            params = copy.deepcopy(item)
            if self.task == 'classification':
                if item['estimator'] == 'decision_tree':
                    params.pop('estimator')
                    pipe = Pipeline(steps=[('imputer',  Base.CustomImputer()),
                                    ('dummie', Base.GetDummiesTransformer()),
                                    ('estimator', Base.Estimator(DecisionTreeClassifier(**params)))])

                elif item['estimator'] == 'xgb':
                    params.pop('estimator')
                    pipe = Pipeline(steps=[('imputer',  Base.CustomImputer()),
                                            ('dummie', Base.GetDummiesTransformer()),
                                            ('estimator', Base.Estimator(XGBClassifier(**params)))])

                elif item['estimator'] == 'catboost':
                    params.pop('estimator')
                    pipe = Pipeline(steps=[('imputer',Base.CustomImputer()),
                                            ('estimator', Base.Estimator(CatBoostClassifierScikit(params)))])

                elif item['estimator'] == 'adaboost':
                    params.pop('estimator')                         
                    pipe = Pipeline(steps=[('imputer',  Base.CustomImputer()),
                                            ('dummie', Base.GetDummiesTransformer()),
                                            ('estimator', Base.Estimator(AdaBoostClassifier(**params)))])

                elif item['estimator'] == 'lgbm':
                    params.pop('estimator')
                    pipe = Pipeline(steps=[('imputer',  Base.CustomImputer()),
                                            ('dummie', Base.GetDummiesTransformer()),
                                            ('estimator', Base.Estimator(LGBMClassifier(**params)))])
                elif item['estimator'] == 'logistic_regression':
                    params.pop('estimator')
                    pipe = Pipeline(steps=[('imputer',Base.CustomImputer() ),
                                        ('dummie', Base.GetDummiesTransformer()),
                                        ('estimator', Base.Estimator(LogisticRegression(**params)))])
                else:
                    raise BaseException(
                        "This model in not implemented yet")
            elif self.task == 'regression':
                if item['estimator'] == 'xgb':
                    params.pop('estimator')
                    pipe = Pipeline(steps=[('imputer', Base.CustomImputer()),
                                        ('dummie', Base.GetDummiesTransformer()),
                                        ('estimator', Base.Estimator(XGBRegressor(**params)))])

                elif item['estimator'] == 'decision_tree':
                    params.pop('estimator')
                    pipe = Pipeline(steps=[('imputer', Base.CustomImputer()),
                                        ('dummie', Base.GetDummiesTransformer()),
                                        ('estimator', Base.Estimator(DecisionTreeRegressor(**params)))])

                elif item['estimator'] == 'catboost':
                    params.pop('estimator')
                    pipe = Pipeline(steps=[('imputer',  Base.CustomImputer()),
                                        ('estimator', Base.Estimator(CatBoostRegressorScikit(params)))])

                elif item['estimator'] == 'adaboost':
                    params.pop('estimator')
                    pipe = Pipeline(steps=[('imputer',  Base.CustomImputer()),
                                            ('dummie', Base.GetDummiesTransformer()),
                                            ('estimator', Base.Estimator(AdaBoostRegressor(**params)))])

                elif item['estimator'] == 'lgbm':
                    params.pop('estimator')
                    pipe = Pipeline(steps=[('imputer',  Base.CustomImputer()),
                                            ('dummie', Base.GetDummiesTransformer()),
                                            ('estimator', Base.Estimator(LGBMRegressor(**params)))])

                elif item['estimator'] == 'svr':
                    params.pop('estimator')                     
                    pipe = Pipeline(steps=[('imputer',  Base.CustomImputer()),
                                        ('dummie', Base.GetDummiesTransformer()),
                                        ('estimator', Base.Estimator(SVR(**params)))])
                else:
                    raise BaseException(
                        "This model in not implemented yet")
            models.append(pipe.fit(X_train, y_train))
        return models

    def run_pipe_with_fs(self, model_params, fs_params, X_train, y_train):
        '''Creates, Run, and stores a full pipe with feauture selection for each possible parameter setting'''
        models = []
        for i, fs_settings in enumerate(fs_params):
            for j, item in enumerate(model_params):
                print("Fit ", (i+1)*(j+1), "/ " + str(len(model_params)*len(fs_params)))
                params = copy.deepcopy(item)
                if self.task == 'classification':
                    if item['estimator'] == 'decision_tree':
                        params.pop('estimator')
                        pipe = Pipeline(steps=[('imputer', Base.CustomImputer()),
                                                ('dummie', Base.GetDummiesTransformer()),
                                                ('fs', Base.FeatureSelector(fs_settings)),
                                                ('estimator', Base.Estimator(DecisionTreeClassifier(**params)))])
                    elif item['estimator'] == 'xgb':
                        params.pop('estimator')
                        pipe = Pipeline(steps=[('imputer', Base.CustomImputer()),
                                                ('dummie', Base.GetDummiesTransformer()),
                                                ('fs', Base.FeatureSelector(fs_settings)),
                                                ('estimator', Base.Estimator(XGBClassifier(**params)))])
                    elif item['estimator'] == 'catboost':
                        params.pop('estimator')
                        pipe = Pipeline(steps=[('imputer',Base.CustomImputer()),
                                                ('fs', Base.FeatureSelector(fs_settings)),
                                                ('estimator', Base.Estimator(CatBoostClassifierScikit(params)))])
                    elif item['estimator'] == 'adaboost':
                        params.pop('estimator')
                        pipe = Pipeline(steps=[('imputer',Base.CustomImputer()),
                                                ('dummie', Base.GetDummiesTransformer()),
                                                ('fs', Base.FeatureSelector(fs_settings)),
                                                ('estimator', Base.Estimator(AdaBoostClassifier(**params)))])
                    elif item['estimator'] == 'lgbm':
                        params.pop('estimator')
                        pipe = Pipeline(steps=[('imputer',Base.CustomImputer()),
                                                ('dummie', Base.GetDummiesTransformer()),
                                                ('fs', Base.FeatureSelector( fs_settings)),
                                                ('estimator', Base.Estimator(LGBMClassifier(**params)))])
                    elif item['estimator'] == 'logistic_regression':
                        params.pop('estimator')
                        pipe = Pipeline(steps=[('imputer',Base.CustomImputer()),
                                                ('dummie', Base.GetDummiesTransformer()),
                                                ('fs', Base.FeatureSelector(fs_settings)),
                                                ('estimator', Base.Estimator(LogisticRegression(**params)))])
                    else:
                        raise BaseException(
                            "This model in not implemented yet")
                elif self.task == 'regression':
                    if item['estimator'] == 'xgb':
                        params.pop('estimator')
                        pipe = Pipeline(steps=[('imputer', Base.CustomImputer()),
                                                ('dummie', Base.GetDummiesTransformer()),
                                                ('fs', Base.FeatureSelector(fs_settings)),
                                                ('estimator', Base.Estimator(XGBRegressor(**params)))])
                    elif item['estimator'] == 'decision_tree':
                        params.pop('estimator')
                        pipe = Pipeline(steps=[('imputer', Base.CustomImputer()),
                                                ('dummie', Base.GetDummiesTransformer()),
                                                ('fs', Base.FeatureSelector(fs_settings)),
                                                ('estimator', Base.Estimator(DecisionTreeRegressor(**params)))])
                    elif item['estimator'] == 'catboost':
                        params.pop('estimator')
                        pipe = Pipeline(steps=[('imputer', Base.CustomImputer()),
                                                ('fs', Base.FeatureSelector(fs_settings)),
                                                ('estimator', Base.Estimator(CatBoostRegressorScikit(params)))])
                    elif item['estimator'] == 'adaboost':
                        params.pop('estimator')
                        pipe = Pipeline(steps=[('imputer',Base.CustomImputer()),
                                                ('dummie', Base.GetDummiesTransformer()),
                                                ('fs', Base.FeatureSelector( fs_settings)),
                                                ('estimator', Base.Estimator(AdaBoostRegressor(**params)))])
                    elif item['estimator'] == 'lgbm':
                        params.pop('estimator')
                        pipe = Pipeline(steps=[('imputer', Base.CustomImputer()),
                                                ('dummie', Base.GetDummiesTransformer()),
                                                ('fs', Base.FeatureSelector(fs_settings)),
                                                ('estimator', Base.Estimator(LGBMRegressor(**params)))])
                    elif item['estimator'] == 'svr':
                        params.pop('estimator')
                        pipe = Pipeline(steps=[('imputer', Base.CustomImputer()),
                                                ('dummie', Base.GetDummiesTransformer()),
                                                ('fs', Base.FeatureSelector( fs_settings)),
                                                   ('estimator', Base.Estimator(SVR(**params)))])
                    else:
                        raise BaseException(
                            "This model in not implemented yet")
                pipe.fit(X_train, y_train)
                models.append(pipe)
        return models
    def __get_best_model__(self, task, models, X_test, y_test):
        '''
            Finds the best model among the models returned by grid search"
        '''
        best_result_classification = 0
        best_result_regresion = None
        y_pred = models[0].predict(X_test)
        if task == 'classification':
            best_model = {'model': models[0],
                          'auc': best_result_classification}
        elif task == 'regression':
            best_result_regresion = Metrics.__theil_stat__(y_test, y_pred)
            best_model = {'model': models[0],
                          'theil_stat': best_result_regresion}

        for model in models:
            y_pred = model.predict(X_test)
            if task == 'classification':
                if not self.compute_ks:
                    value = None
                    if len(np.unique(y_test)) > 2:
                        y_pred = model.predict_proba(X_test)           
                        value = roc_auc_score(y_test, y_pred,multi_class= 'ovo',   average='weighted')
                    else:
                        value = roc_auc_score(y_test, y_pred)
                    if value > best_result_classification:
                        best_model = {'model': model, 'auc': value}
                        best_result_classification = value
                else:
                    return self.get_best_model_by_ks(models, X_test, y_test)
            elif task == 'regression':
                value = Metrics.__theil_stat__(y_test, y_pred)
                if value < best_result_regresion:
                    best_model = {'model': model, 'theil_stat': value}
                    best_result_regresion = value
            print('Best score ', best_model[list(best_model.keys())[1]])
        return best_model['model']

    def get_best_model_by_ks(self, models, X, y):

        num_models_accepted = int(np.ceil(self.acceptance_rate*len(models)))
        best_models = ['']*(num_models_accepted)

        aucs = []

        for model in models:
            y_pred = model.predict(X)
            auc = roc_auc_score(y, y_pred)
            aucs.append((auc, model))

        aucs.sort(key=lambda x: x[0], reverse=True)
        best_models = aucs[:num_models_accepted]

        model_ks = []
        for item in best_models:
            y_pred = item[1].predict(X)
            ks = Metrics.__compute_ks_score__(y, y_pred)
            model_ks.append((ks, item[1]))

        model_ks.sort(key=lambda x: x[0], reverse=True)
        print('The model KS is: ', model_ks[0][0])
        return model_ks[0][1]



import pandas as pd
from pandas.api.types import is_numeric_dtype
import numpy as np

from sklearn.metrics import roc_auc_score
from sklearn.ensemble import AdaBoostClassifier, AdaBoostRegressor
from sklearn.tree import DecisionTreeClassifier, DecisionTreeRegressor
from sklearn.linear_model import LogisticRegression
from sklearn.svm import SVR

from xgboost import XGBClassifier, XGBRegressor
from catboost import CatBoostClassifier, CatBoostRegressor, Pool
from lightgbm import LGBMClassifier, LGBMRegressor

from paml.metrics import Metrics


class OptunaObjectives:
    def __objective_xgb_clf__(trial, X, y):
        params = {           
            'use_label_encoder': False,
          
            'n_estimators': 1000,
            'eval_metric': ["auc"],
            'seed': 42,
            'learning_rate': trial.suggest_loguniform('learning_rate', 1e-5, 1e-2),
            'gamma': trial.suggest_uniform('gamma', 0, 1),
            'max_depth': trial.suggest_int('max_depth', 3, 10),
            'min_child_weight': trial.suggest_int('min_child_weight', 0, 10),
            'subsample': trial.suggest_uniform('subsample', 0.5, 1),
            'colsample_bytree': trial.suggest_uniform('colsample_bytree', 0, 1),
            'alpha': trial.suggest_loguniform('alpha', 1e-8, 10.0),
            'lambda': trial.suggest_loguniform('lambda', 1e-8, 10.0)
        }
        xgbClassifier = XGBClassifier(**params)

        try:
            # numeric_features = [col for col in X.columns if X[col].dtypes == 'int64' or X[col].dtypes == 'float64' 
            #                     or X[col].dtypes == 'int' or X[col].dtypes == 'float']
            numeric_features = [col for col in X.columns if is_numeric_dtype(X[col]) or np.all(X[col].str.isnumeric())]
            if numeric_features:
                for feat in numeric_features.copy():
                    X.loc[:][feat] = X[feat].astype('float64')
        except:
            pass

        X = pd.get_dummies(X, drop_first = True)
        xgbClassifier.fit(X, y)
        y_pred = None
        roc = None
        if len(np.unique(y) > 2):
            y_pred = xgbClassifier.predict_proba(X)
            roc = roc_auc_score(y, y_pred, multi_class='ovr', average='weighted')
        else:
            y_pred = xgbClassifier.predict(X)
            roc = roc_auc_score(y, y_pred)
        return roc

    def __objective_xgb_regressor__(trial, X, y):
        params = {
            'n_estimators': 1000,
            'eval_metric': trial.suggest_categorical("eval_metric", ["rmse"]),
            'seed': 42,
            'learning_rate': trial.suggest_loguniform('learning_rate', 1e-5, 1e-2),
            'gamma': trial.suggest_uniform('gamma', 0, 1),
            'max_depth': trial.suggest_int('max_depth', 3, 10),
            'min_child_weight': trial.suggest_int('min_child_weight', 0, 10),
            'subsample': trial.suggest_uniform('subsample', 0.5, 1),
            'colsample_bytree': trial.suggest_uniform('colsample_bytree', 0, 1),
            'alpha': trial.suggest_loguniform('alpha', 1e-8, 10.0),
            'lambda': trial.suggest_loguniform('lambda', 1e-8, 10.0)
        }
        xgbRegressor = XGBRegressor(**params)

        try:
            numeric_features = [col for col in X.columns if is_numeric_dtype(X[col]) or np.all(X[col].str.isnumeric())]
                   
            if numeric_features:
                for feat in numeric_features:
                    X.loc[:][feat] = X[feat].astype('float64')
        except:
            pass

        X = pd.get_dummies(X, drop_first = True)
        xgbRegressor.fit(X, y)
        y_pred = xgbRegressor.predict(X)
        return Metrics.__theil_stat__(y, y_pred)

    def __objective_catboost_classifier__(trial, X, y):
        params = {
            'n_estimators': trial.suggest_int('n_estimators',  200, 500),
            'early_stopping_rounds': 50,
            'eval_metric': trial.suggest_categorical('eval_metric', ['AUC']),
            'random_state': 42,
            'learning_rate': trial.suggest_loguniform('learning_rate', 1e-5, 1e-2),
            'max_depth': trial.suggest_int('max_depth',  3, 8),
            'min_data_in_leaf': trial.suggest_int('min_data_in_leaf', 10, 100),
            'colsample_bylevel': trial.suggest_uniform('colsample_bylevel', 0.5, 1),
            'boosting_type': trial.suggest_categorical('boosting_type', ['Plain', 'Ordered']),
            'l2_leaf_reg': trial.suggest_int('l2_leaf_reg', 1, 10),
            'loss_function': trial.suggest_categorical('loss_function', ['Logloss']),
            'bootstrap_type': trial.suggest_categorical('bootstrap_type', ['Bayesian', 'Bernoulli', 'MVS']),
            
        }
        catboostClassifier = CatBoostClassifier(**params)

        try:
            numeric_features = [col for col in X.columns if is_numeric_dtype(X[col]) or np.all(X[col].str.isnumeric())]
            if numeric_features:
                for feat in numeric_features:
                    # X.loc[:,X.columns == feat] =X.loc[:,X.columns == feat].astype('float64')
                    X.loc[:][feat] = X[feat].astype('float64')

        except:
            pass

        cat_features_index = [X.columns.get_loc(
            x) for x in X.columns if X[x].dtypes == 'object']

        pool = Pool(X, label=y, cat_features=cat_features_index)
        catboostClassifier.fit(pool)


        y_pred = catboostClassifier.predict(pool)
        roc_score = roc_auc_score(y, y_pred,)
   

        return roc_score


    def __objective_catboost_classifier_multi__(trial, X, y):
        params = {
            'n_estimators': trial.suggest_int('n_estimators',  200, 500),
            'early_stopping_rounds': 50,
            'eval_metric': trial.suggest_categorical('eval_metric', ['MultiClass']),
            'random_state': 42,
            'learning_rate': trial.suggest_loguniform('learning_rate', 1e-5, 1e-2),
            'max_depth': trial.suggest_int('max_depth',  3, 8),
            'min_data_in_leaf': trial.suggest_int('min_data_in_leaf', 10, 100),
            'colsample_bylevel': trial.suggest_uniform('colsample_bylevel', 0.5, 1),
            'boosting_type': trial.suggest_categorical('boosting_type', ['Plain', 'Ordered']),
            'l2_leaf_reg': trial.suggest_int('l2_leaf_reg', 1, 10),
            'loss_function': trial.suggest_categorical('loss_function', ['MultiClass']),
            'bootstrap_type': trial.suggest_categorical('bootstrap_type', ['Bayesian', 'Bernoulli', 'MVS']),
            
        }
        
        catboostClassifier = CatBoostClassifier(**params)
        try:
            numeric_features = [col for col in X.columns if is_numeric_dtype(X[col]) or np.all(X[col].str.isnumeric())]
            if numeric_features:
                for feat in numeric_features:
                    X.loc[:][feat] = X[feat].astype('float64')
        except:
            pass

        cat_features_index = [X.columns.get_loc(
            x) for x in X.columns if X[x].dtypes == 'object']

        pool = Pool(X, label=y, cat_features=cat_features_index)
        catboostClassifier.fit(pool)

 
        y_pred = catboostClassifier.predict_proba(pool)
        roc_score = roc_auc_score(y, y_pred,multi_class= 'ovo',   average='weighted')

        return roc_score

    def __objective_catboost_regressor__(trial, X, y):
        params = {
            'n_estimators': 1000,
            'early_stopping_rounds': 50,
            'eval_metric': trial.suggest_categorical('eval_metric', ['RMSE']),
            'random_state': 42,
            'learning_rate': trial.suggest_loguniform('learning_rate', 1e-5, 1e-2),
            'max_depth': trial.suggest_int('max_depth',  3, 8),
            'min_data_in_leaf': trial.suggest_int('min_data_in_leaf', 10, 100),
            'colsample_bylevel': trial.suggest_uniform('colsample_bylevel', 0.5, 1),
            'boosting_type': trial.suggest_categorical('boosting_type', ['Plain', 'Ordered']),
            'l2_leaf_reg': trial.suggest_int('l2_leaf_reg', 1, 10),
            'loss_function': trial.suggest_categorical('loss_function', ['RMSE']),
            'bootstrap_type': trial.suggest_categorical('bootstrap_type', ['Bayesian', 'Bernoulli', 'MVS'])
        }
        catboostRegressor = CatBoostRegressor(**params)
        try:
            numeric_features = [col for col in X.columns if is_numeric_dtype(X[col]) or np.all(X[col].str.isnumeric())]
            if numeric_features:
                for feat in numeric_features:
                    X.loc[:][feat] = X[feat].astype('float64')
        except:
            pass

        cat_features_index = [X.columns.get_loc(
            x) for x in X.columns if X[x].dtypes == 'object']

        pool = Pool(X, label=y, cat_features=cat_features_index)

        catboostRegressor.fit(pool)

        y_pred = catboostRegressor.predict(pool)
        return Metrics.__theil_stat__(y, y_pred)

    def __objective_lgbm_classifier__(trial, X, y):
        params = {
            'objective': 'binary',
            'n_estimators': 1500,
            'eval_metric': trial.suggest_categorical('metric', ['AUC']),
            'seed': 42,
            'boosting': trial.suggest_categorical('boosting', ['gbdt']),
            'num_leaves': trial.suggest_int('num_leaves', 2, 256),
            'learning_rate': trial.suggest_loguniform('learning_rate', 1e-5, 1e-2),
            'max_depth': trial.suggest_int('max_depth', 0, 7),
            'min_data_in_leaf': trial.suggest_int('min_data_in_leaf', 10, 100),
            'min_sum_hessian_in_leaf': trial.suggest_loguniform('min_sum_hessian_in_leaf', 1e-5, 1e-1),
            'reg_alpha': trial.suggest_loguniform('reg_alpha', 1e-8, 10.0),
            'reg_lambda': trial.suggest_loguniform('reg_lambda', 1e-8, 10.0),
            'bagging_freq': 5,
            'bagging_fraction': trial.suggest_uniform('bagging_fraction', 0.1, 1.0),
            'feature_fraction': trial.suggest_uniform('feature_fraction', 0.4, 1.0),
            'feature_fraction_bynode': trial.suggest_uniform('feature_fraction', 0.4, 1.0),
            'min_data_per_group': trial.suggest_int('min_data_per_group', 50, 300),
            'cat_l2': trial.suggest_int('cat_l2', 1, 100),
            'cat_smooth': trial.suggest_int('cat_smooth', 1, 100)
        }

        LGBMClf = LGBMClassifier(**params)

        try:
            numeric_features = [col for col in X.columns if is_numeric_dtype(X[col]) or np.all(X[col].str.isnumeric())]
            if numeric_features:
                for feat in numeric_features:
                    X.loc[:][feat] = X[feat].astype('float64')
        except:
            pass

        X = pd.get_dummies(X, drop_first=True)
        LGBMClf.fit(X, y, eval_set=[(X, y)])
        y_pred = LGBMClf.predict(X)
        return roc_auc_score(y, y_pred)

    def __objective_lgbm_classifier_multi__(trial, X, y):
        params = {
            'objective': 'multiclass',
            'n_estimators': 1500,
            'eval_metric': trial.suggest_categorical('metric', ['multiclass']),
            'seed': 42,
            'boosting': trial.suggest_categorical('boosting', ['gbdt']),
            'num_leaves': trial.suggest_int('num_leaves', 2, 256),
            'learning_rate': trial.suggest_loguniform('learning_rate', 1e-5, 1e-2),
            'max_depth': trial.suggest_int('max_depth', 0, 7),
            'min_data_in_leaf': trial.suggest_int('min_data_in_leaf', 10, 100),
            'min_sum_hessian_in_leaf': trial.suggest_loguniform('min_sum_hessian_in_leaf', 1e-5, 1e-1),
            'reg_alpha': trial.suggest_loguniform('reg_alpha', 1e-8, 10.0),
            'reg_lambda': trial.suggest_loguniform('reg_lambda', 1e-8, 10.0),
            'bagging_freq': 5,
            'bagging_fraction': trial.suggest_uniform('bagging_fraction', 0.1, 1.0),
            'feature_fraction': trial.suggest_uniform('feature_fraction', 0.4, 1.0),
            'feature_fraction_bynode': trial.suggest_uniform('feature_fraction', 0.4, 1.0),
            'min_data_per_group': trial.suggest_int('min_data_per_group', 50, 300),
            'cat_l2': trial.suggest_int('cat_l2', 1, 100),
            'cat_smooth': trial.suggest_int('cat_smooth', 1, 100)
        }

        LGBMClf = LGBMClassifier(**params)

        try:
            numeric_features = [col for col in X.columns if is_numeric_dtype(X[col]) or np.all(X[col].str.isnumeric())]
            if numeric_features:
                for feat in numeric_features:
                    X.loc[:][feat] = X[feat].astype('float64')
        except:
            pass

        X = pd.get_dummies(X, drop_first=True)
        LGBMClf.fit(X, y, eval_set=[(X, y)])
        y_pred = LGBMClf.predict_proba(X)
        roc_auc = roc_auc_score(y, y_pred, average='weighted', multi_class='ovr')
        
        return roc_auc

    def __objective_lgbm_regressor__(trial, X, y):
        params = {
            'objective': 'binary',
            'n_estimators': 1500,
            'eval_metric': trial.suggest_categorical('metric', ['rmse']),
            'seed': 42,
            'boosting': trial.suggest_categorical('boosting', ['gbdt']),
            'num_leaves': trial.suggest_int('num_leaves', 2, 256),
            'learning_rate': trial.suggest_loguniform('learning_rate', 1e-5, 1e-1),
            'max_depth': trial.suggest_int('max_depth', 0, 7),
            'min_data_in_leaf': trial.suggest_int('min_data_in_leaf', 10, 100),
            'min_sum_hessian_in_leaf': trial.suggest_loguniform('min_sum_hessian_in_leaf', 1e-5, 1e-2),
            'reg_alpha': trial.suggest_loguniform('reg_alpha', 1e-8, 10.0),
            'reg_lambda': trial.suggest_loguniform('reg_lambda', 1e-8, 10.0),
            'bagging_freq': 5,
            'bagging_fraction': trial.suggest_uniform('bagging_fraction', 0.1, 1.0),
            'feature_fraction': trial.suggest_uniform('feature_fraction', 0.4, 1.0),
            'feature_fraction_bynode': trial.suggest_uniform('feature_fraction', 0.4, 1.0),
            'min_data_per_group': trial.suggest_int('min_data_per_group', 50, 300),
            'cat_l2': trial.suggest_int('cat_l2', 1, 100),
            'cat_smooth': trial.suggest_int('cat_smooth', 1, 100)
        }

        LGBMRgs = LGBMRegressor(**params)

        try:
            numeric_features = [col for col in X.columns if is_numeric_dtype(X[col]) or np.all(X[col].str.isnumeric())]
            if numeric_features:
                for feat in numeric_features:
                    X.loc[:][feat] = X[feat].astype('float64')
        except:
            pass

        X = pd.get_dummies(X, drop_first=True)
        LGBMRgs.fit(X, y, eval_set=[(X, y)])
        y_pred = LGBMRgs.predict(X)
        return Metrics.__theil_stat__(y, y_pred)

    def __objective_decision_tree_classifier__(trial, X, y):
        params = {'criterion': trial.suggest_categorical('criterion', ['gini', 'entropy']),
                  'splitter': trial.suggest_categorical('splitter', ['best', 'random']),
                  'max_depth': trial.suggest_int('max_depth', 2, 12),
                  'max_features': trial.suggest_categorical('max_features', ['auto', 'sqrt', 'log2']),
                  'random_state': 42}
        DTClassifier = DecisionTreeClassifier(**params)

        try:
            numeric_features = [col for col in X.columns if is_numeric_dtype(X[col]) or np.all(X[col].str.isnumeric())]
            if numeric_features:
                for feat in numeric_features:
                    X.loc[:][feat] = X[feat].astype('float64')
        except:
            pass

        X = pd.get_dummies(X, drop_first=True)
        DTClassifier.fit(X, y)

        y_pred = None
        roc = None
        if len(np.unique(y)) > 2:
            y_pred = DTClassifier.predict_proba(X)
            roc = roc_auc_score(y, y_pred, multi_class='ovr', average='weighted')
        else:
            y_pred = DTClassifier.predict(X)
            roc = roc_auc_score(y, y_pred)
        return roc

    def __objective_decision_tree_regressor__(trial, X, y):
        params = {'criterion': trial.suggest_categorical('criterion', ['mse', 'friedman_mse', 'poisson']),
                  'splitter': trial.suggest_categorical('splitter', ['best', 'random']),
                  'max_depth': trial.suggest_int('max_depth', 2,  12),
                  'max_features': trial.suggest_categorical('max_features', ['auto', 'sqrt', 'log2']),
                  'random_state': 42}
        DTRegressor = DecisionTreeRegressor(**params)


        try:
            numeric_features = [col for col in X.columns if is_numeric_dtype(X[col]) or np.all(X[col].str.isnumeric())]
            if numeric_features:
                for feat in numeric_features:
                    X.loc[:][feat] = X[feat].astype('float64')
        except:
            pass


        X = pd.get_dummies(X, drop_first=True)
        DTRegressor.fit(X, y)

        y_pred = DTRegressor.predict(X)
        return Metrics.__theil_stat__(y, y_pred)

    def __objective_adaboost_classifier__(trial, X, y):
        params = {'n_estimators': trial.suggest_int('n_estimators', 500, 1500),
                  'algorithm': trial.suggest_categorical('algorithm', ["SAMME", "SAMME.R"]),
                  'learning_rate': trial.suggest_uniform('learning_rate', 1e-2,  1), }

        adaboost_clf = AdaBoostClassifier(**params)

        try:
            numeric_features = [col for col in X.columns if is_numeric_dtype(X[col]) or np.all(X[col].str.isnumeric())]
            if numeric_features:
                for feat in numeric_features:
                    X.loc[:][feat] = X[feat].astype('float64')
        except:
            pass

        X = pd.get_dummies(X, drop_first=True)
        adaboost_clf.fit(X, y)
        y_pred = None
        roc = None
        if len(np.unique(y)) > 2:
            y_pred = adaboost_clf.predict_proba(X)
            roc = roc_auc_score(y, y_pred, multi_class='ovr', average='weighted')
        else:
            y_pred = adaboost_clf.predict(X)
            roc = roc_auc_score(y, y_pred)
        return roc



    def __objective_adaboost_regressor__(trial, X, y):
        params = {'n_estimators': trial.suggest_int('n_estimators', 500, 1500),
                  'loss': trial.suggest_categorical('loss', ["linear", "square", "exponential"]),
                  'learning_rate': trial.suggest_uniform('learning_rate', 1e-2,  1), }

        adaboostRegressor = AdaBoostRegressor(**params)

        try:
            numeric_features = [col for col in X.columns if is_numeric_dtype(X[col]) or np.all(X[col].str.isnumeric())]
            if numeric_features:
                for feat in numeric_features:
                    X.loc[:][feat] = X[feat].astype('float64')
        except:
            pass

        X = pd.get_dummies(X, drop_first=True)
        adaboostRegressor.fit(X, y)
        y_pred = adaboostRegressor.predict(X)
        return Metrics.__theil_stat__(y, y_pred)

    def __objective__logistic_regression__(trial, X, y):
        params = {
                'solver': trial.suggest_categorical('solver', ['lbfgs', 'liblinear', 'sag', 'saga']),
                'C': trial.suggest_uniform('C', 0.3, 1)}
        logistic = LogisticRegression(**params)

        try:
            numeric_features = [col for col in X.columns if is_numeric_dtype(X[col]) or np.all(X[col].str.isnumeric())]
            if numeric_features:
                for feat in numeric_features:
                    X.loc[:][feat] = X[:][feat].astype('float64')
        except:
            pass


        X = pd.get_dummies(X, drop_first=True)
        logistic.fit(X, y)
        y_pred = None
        roc = None
        if len(np.unique(y)) > 2:
            y_pred = logistic.predict_proba(X)
            roc = roc_auc_score(y, y_pred, multi_class='ovr', average='weighted')
        else:
            y_pred = logistic.predict(X)
            roc = roc_auc_score(y, y_pred)
        return roc
      

    def __objective_svr_regressor__(trial, X, y):
        params = {'kernel': trial.suggest_categorical('kernel', ["linear", "poly", "sigmoid"]),
                  'gamma': trial.suggest_categorical('gamma', ["scale", "auto"]),
                  'C': trial.suggest_uniform('C', 0.2, 1),
                  'epsilon': trial.suggest_uniform('epsilon', 1e-2, 1)}

        svr = SVR(**params)

        try:
            numeric_features = [col for col in X.columns if is_numeric_dtype(X[col]) or np.all(X[col].str.isnumeric())]
            if numeric_features:
                for feat in numeric_features:
                    X.loc[:][feat] = X[feat].copy().astype('float64')
        except:
            pass

        X = pd.get_dummies(X, drop_first=True)
        svr.fit(X, y)
        y_pred = svr.predict(X)
        return Metrics.__theil_stat__(y, y_pred)



from paml.searches import OptunaSearch
from paml.searches import GridSearchCV
from paml.searches import GridSearch

import numpy as np
class AutoML:
    ''''
    Parameters:
        task - Machine Learning task, it can be ‘regression’ or ‘classification’
        
        search_space: list Default None - The parameter search_space for one or more models.If not defined by 
        the user the  software will provide a default space.
        
        search: string Default: Optuna -  The search type to be performed by the autoML. The available options are ‘GridSearch’, ‘GridSearchCV’ and ‘OptunaSearch’. When using the ‘OptunaSearch’ is not possible to define a search_space

        models: list: Default: ['all'] - The model(s) that will be used in the autoML. It is mandatory to pass a list with the models, even if there is only one model. If option ['all'] is passed the tool will use all the models implemented. Available models are: DecisionTreeClassifier, DecisionTreeRegressor,  CatBoostClassifier, CatBoostRegressor, LGBMClasssifier, LGBMRegressor, XGBoostClassifier, XGBoostRgressor, AdaBoostClassifier, AdaBoostRegressor, LogisticRegression and SVR.  When passing one of these models use one of the following alias: decision_tree, catboost, lgbm, xgb, adaboost, logistic_regression, svr or all.

        compute_ks: bool: Default: False - Boolean flag for using the Kolmogorov-Smirnov (KS) score as measure for choosing between the best ranked AUC scored models. It is only used in classification tasks.

        n_folds: int: Default: 5 - Number of folds used in cross-validation. It is only used in GridSearchCV and OptunaSearch.

        feature_selection: bool: Deafult: False - Boolean flag for performing or not feature selection on the dataset.

        fs_params: dict: Default: {'max_depth': [5]} - Dict of parameterization of random forest estimator used in feature selection. Only used when feature_selection is True.

        acceptance_rate: float: Default 0.001 - percentage of number of models acceptation of all possible combinations trained by the searched. It is used when the flag compute_ks is set to True. From the 0.1% best ranked models, the model with best KS is chosen.

        n_trials : int Deafult: 10 - Number of trials for the Optuna search.

        n_jobs: int Default: - 1 Number of jobs to run in parallel. -1 means using all processors

    '''
    def __init__(self,  task: str, search_space=None, search: str = 'OptunaSearch', models=['all'],
                 compute_ks: bool = False, n_folds: int = 5, feature_selection=False, fs_params={'max_depth': [5]},
                 acceptance_rate: float = 0.01,
                 n_trials: int = 10,
                 n_jobs: int = -1):

        self.search = search
        self.task = task
        self.models = models
        self.compute_ks = compute_ks
        self.search_space = search_space
        self.n_folds = n_folds
        self.acceptance_rate = acceptance_rate
        self.n_trials = n_trials
        self.feature_selection = feature_selection
        self.n_jobs = n_jobs
        self.fs_params = fs_params

    def fit(self, X, y):

        if self.search == 'GridSearchCV':
            gridSearchCV = GridSearchCV(search_space=self.search_space, n_jobs=self.n_jobs,
                                        n_folds=self.n_folds, models=self.models, task=self.task,
                                        compute_ks=self.compute_ks, feature_selection=self.feature_selection,
                                        acceptance_rate=self.acceptance_rate, fs_params={'max_depth': 5})

            best_model = gridSearchCV.fit_model(X, y)
            return best_model
        elif self.search == 'GridSearch':
            gridSearch = GridSearch(search_space=self.search_space,
                                    models=self.models, task=self.task,
                                    compute_ks=self.compute_ks, acceptance_rate=self.acceptance_rate,
                                    feature_selection=self.feature_selection)
            model = gridSearch.fit(X, y)
            return model

        elif self.search == 'OptunaSearch':
            if self.search_space:
                print("You can't use a search space when using optuna")
            multi_class = False
            if len(np.unique(y)) > 2 and self.task == 'classification' :
                multi_class = True

            Optuna = OptunaSearch(n_trials=self.n_trials, task=self.task,
                                  compute_ks=self.compute_ks, n_folds=self.n_folds,
                                  models=self.models,
                                  feature_selection=self.feature_selection, multi_class=multi_class)
            best_model = Optuna.fit(X, y)
            return best_model
        else:
            raise BaseException("Search not implemented yet")



from sklearn.metrics import make_scorer
import pandas as pd

import numpy as np

class Metrics:
    def __theil_stat__(y_true, y_pred):
        ''' Computes theils bias statistics'''
        return (np.mean(y_pred) - np.mean(np.array(y_true)))/np.mean(((y_pred - np.array(y_true)))**2)

    def theil_stat():
        '''Computes make scorer for theil bias statistics'''
        return  make_scorer(score_func= Metrics.__theil_stat__, needs_proba= False, greater_is_better=False)

    def __compute_ks_score__(target, score):
        """ Calculate KS score for provided data """
        b2 = pd.crosstab(score, target, rownames = ['score']).reset_index()
        if len(np.unique(target)) > 1:
            freq_bom = b2.iloc[:,1]/sum(b2.iloc[:,1])
            freq_mau = b2.iloc[:,2]/sum(b2.iloc[:,2])
            ac_bom = np.cumsum(freq_bom)
            ac_mau = np.cumsum(freq_mau)
            diff = np.abs(ac_bom - ac_mau).round(7)
            ks = diff.max() * 100
        else:
            ks = 0      
        return ks


from setuptools import setup

def readme():
    with open('README.rst') as f:
        return f.read()

setup(name='pamel',
      version='0.0.69',
      long_description=readme(),
      long_description_content_type='text/x-rst;',
      classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.8',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
      ],
      keywords='automl, credit',
      author='Rute Souza de Abreu',
      author_email='rute.s.abreu@gmail.com',
      license='MIT',
      packages=['paml'],
      install_requires=[
          'scikit-learn',
          'numpy',
          'Boruta',
          'pandas',
          'xgboost',
          'catboost',
          'lightgbm',
          'optuna'
      ],
      include_package_data=True,
      zip_safe=True)
